import torch

from datetime import timedelta
from timeit import default_timer as timer
from torch.utils.data import DataLoader

from .integrator import Integrator
from eeyore.datasets import DataCounter

class MCIntegrator(Integrator):
    def __init__(self, f=None, samples=None):
        super(MCIntegrator, self).__init__()
        self.f = f
        self.samples = samples

    def integrate(self, x, y):
        integral = 0
        for sample in self.samples:
            integral = integral + self.f(sample, x, y)
        integral = integral / len(self.samples)

        return integral

    def integrate_from_dataset(
        self, dataset, num_points, shuffle=True, dtype=torch.float64, device='cpu', verbose=False, verbose_step=1):
        dataloader = DataLoader(dataset, batch_size=1, shuffle=shuffle)
        counter = DataCounter.from_dataloader(dataloader)

        counter.set_num_epochs(num_points)
        verbose_msg = self.set_verbose_msg(counter)

        integrals = torch.empty(num_points, dtype=dtype, device=device)
        indices = torch.empty(num_points, dtype=torch.int64, device=device)

        for _ in range(counter.num_epochs):
            for _, (x, y, idx) in enumerate(dataloader):
                if counter.idx >= counter.num_iters:
                    break

                if verbose and (((counter.idx+1) % verbose_step) == 0):
                    start_time = timer()

                integrals[counter.idx] = self.integrate(x, y).item()
                indices[counter.idx] = idx.clone().detach()

                if verbose and (((counter.idx+1) % verbose_step) == 0):
                    end_time = timer()
                    print(verbose_msg.format(counter.idx+1, timedelta(seconds=end_time-start_time)))

                counter.increment_idx()

        return integrals, indices

    def set_verbose_msg(self, counter):
        return "Iteration {:" \
            + str(len(str(counter.num_iters))) \
            + "} out of " \
            + str(counter.num_iters) \
            + ", duration {}"



import torch

from .kernel import Kernel

class NormalizedKernel(Kernel):
    """ Base class for normalized kernels """

    def default_density(self):
        raise NotImplementedError

    def set_density(self):
        raise NotImplementedError

    def log_prob(self, state):
        return torch.sum(self.density.log_prob(state))

    def sample(self):
        """ Sample the probability density function """
        return self.density.sample()



import torch

from .kernel import Kernel

class HomogeneousKernel(Kernel):
    """ Base class for function kernels based on distance between inputs """

    def dist(self, x1, x2):
        return torch.norm(x1 - x2, 2)

    def squared_dist(self, x1, x2):
        return self.dist(x1, x2).pow(2)



import itertools
import torch

class Kernel:
    """ Base class for kernels """

    def k(self, x1, x2):
        raise NotImplementedError

    def check_input_dtype(self, x, dtype):
        if not all(t == dtype for t in [element.dtype for element in x]):
            raise ValueError

    def check_inputs_dtype(self, x1, x2, dtype):
        if not all(element.dtype == dtype for element in itertools.chain.from_iterable([x1, x2])):
            raise ValueError

    def check_input_device(self, x, device):
        if not all(t == device for t in [element.device for element in x]):
            raise ValueError

    def check_inputs_device(self, x1, x2, device):
        if not all(element.device == device for element in itertools.chain.from_iterable([x1, x2])):
            raise ValueError

    def symm_K(self, x, check_input=False):
        n = len(x)
        dtype = x[0].dtype
        device = x[0].device

        if check_input:
            self.check_input_dtype(x, dtype)
            self.check_input_device(x, device)

        result = torch.empty([n, n], dtype=dtype, device=device)

        for i in range(n):
            for j in range(n):
                if i <= j:
                    result[i, j] = self.k(x[i], x[j])
                else:
                    result[i, j] = result[j, i]

        return result

    def K(self, x1, x2, check_input=False):
        n1 = len(x1)
        n2 = len(x2)
        dtype = x1[0].dtype
        device = x1[0].device

        if check_input:
            self.check_inputs_dtype(x1, x2, dtype)
            self.check_inputs_device(x1, x2, device)

        result = torch.empty([n1, n2], dtype=dtype, device=device)

        for i in range(n1):
            for j in range(n2):
                    result[i, j] = self.k(x1[i], x2[j])

        return result

    def sum_symm_K(self, x, include_diag=True, check_input=False):
        n = len(x)
        dtype = x[0].dtype
        device = x[0].device

        if check_input:
            self.check_input_dtype(x, dtype)
            self.check_input_device(x, device)

        result = torch.tensor([0], dtype=dtype, device=device)

        if include_diag:
            for i in range(n):
                result = result + self.k(x[i], x[i])

        for i in range(n):
            for j in range(i):
                    result = result + 2 * self.k(x[i], x[j])

        return result

    def sum_K(self, x1, x2, check_input=False):
        n1 = len(x1)
        n2 = len(x2)
        dtype = x1[0].dtype
        device = x1[0].device

        if check_input:
            self.check_inputs_dtype(x1, x2, dtype)
            self.check_inputs_device(x1, x2, device)

        result = torch.tensor([0], dtype=dtype, device=device)

        for i in range(n1):
            for j in range(n2):
                    result = result + self.k(x1[i], x2[j])

        return result



from .bayesian_model import BayesianModel
from .distribution_model import DistributionModel
from .log_target_model import LogTargetModel
from .logistic_regression import LogisticRegression
from .mlp import MLP
from .model import Model
# from .ode_model import ODEModel



class DataCounter:
    def __init__(self, batch_size, sample_size,
        num_epochs=None, num_burnin_epochs=None, num_batches=None, drop_last=False):
        self.set_data_info(batch_size, sample_size, num_batches=num_batches, drop_last=drop_last)
        self.set_epoch_info(num_epochs, num_burnin_epochs)
        self.reset()

    def set_num_batches(self, drop_last=False):
        self.num_batches = self.sample_size / self.batch_size
        if (self.sample_size % self.batch_size != 0) and not drop_last:
            self.num_batches = self.num_batches + 1

    def set_data_info(self, batch_size, sample_size, num_batches=None, drop_last=False):
        self.batch_size = batch_size
        self.sample_size = sample_size
        if num_batches is None:
            self.set_num_batches(drop_last=drop_last)
        else:
            self.num_batches = num_batches

    def set_data_info_from_dataloader(self, dataloader):
        self.set_data_info(dataloader.batch_size, len(dataloader.dataset), num_batches=len(dataloader))

    def set_num_iters(self, num_epochs):
        self.num_epochs = num_epochs
        if self.num_epochs is not None:
            self.num_iters = self.num_epochs * self.num_batches
        else:
            self.num_iters = None

    def set_num_burnin_iters(self, num_burnin_epochs):
        self.num_burnin_epochs = num_burnin_epochs
        if self.num_burnin_epochs is not None:
            self.num_burnin_iters = self.num_burnin_epochs * self.num_batches
        else:
            self.num_burnin_iters = None

    def set_epoch_info(self, num_epochs, num_burnin_epochs):
        self.set_num_iters(num_epochs)
        self.set_num_burnin_iters(num_burnin_epochs)

    def set_num_epochs(self, num_iters):
        self.num_iters = num_iters
        if self.num_iters is not None:
            if (self.num_iters % self.num_batches) == 0:
                self.num_epochs = self.num_iters // self.num_batches
            else:
                self.num_epochs = self.num_iters // self.num_batches + 1
        else:
            self.num_epochs = None

    def set_num_burnin_epochs(self, num_burnin_iters):
        self.num_burnin_iters = num_burnin_iters
        if self.num_burnin_iters is not None:
            if (self.num_burnin_iters % self.num_batches) == 0:
                self.num_burnin_epochs = self.num_burnin_iters // self.num_batches
            else:
                self.num_burnin_epochs = self.num_burnin_iters // self.num_batches + 1
        else:
            self.num_burnin_epochs = None

    def set_iter_info(self, num_iters, num_burnin_iters):
        self.set_num_epochs(self, num_iters)
        self.set_num_burnin_epochs(self, num_burnin_iters)

    @classmethod
    def from_dataloader(selfclass, dataloader, num_epochs=None, num_burnin_epochs=None):
        return selfclass(
            dataloader.batch_size,
            len(dataloader.dataset),
            num_epochs=num_epochs,
            num_burnin_epochs=num_burnin_epochs,
            num_batches=len(dataloader)
        )

    def reset(self):
        self.idx = 0

    def increment_idx(self, incr=1):
        self.idx = self.idx + incr



import torch

from .single_chain_serial_sampler import SingleChainSerialSampler
from eeyore.chains import ChainList
from eeyore.datasets import DataCounter
from eeyore.tuners import HMCDATuner

class HMC(SingleChainSerialSampler):
    def __init__(self, model,
        theta0=None, dataloader=None, data0=None, counter=None,
        step=0.1, num_steps=10, tuner=None, transform=None, chain=ChainList()):
        super(HMC, self).__init__(counter or DataCounter.from_dataloader(dataloader))
        self.model = model
        self.dataloader = dataloader
        self.tuner = tuner

        if self.tuner is not None:
            if isinstance(self.tuner, HMCDATuner):
                if self.tuner.e0 is None:
                    self.init_step(theta0.clone().detach())
                    if self.tuner.eub is not None:
                        self.step = min(self.tuner.eub, self.step)
                    self.tuner.set_m(self.step)
                else:
                    self.step = self.tuner.e0

                self.num_steps = self.tuner.num_steps(self.step)
        else:
            self.step = step
            self.num_steps = num_steps

        self.transform = transform

        self.keys = ['sample', 'target_val', 'grad_val', 'momentum', 'hamiltonian', 'accepted']
        self.chain = chain

        if theta0 is not None:
            self.set_current(theta0.clone().detach(), data=data0)

    def init_step(self, theta):
        iterator = iter(self.dataloader)
        x, y = next(iterator)

        self.step = 1.
        self.num_steps = 1

        current = {
            'sample': theta,
            'momentum': torch.randn(self.model.num_params(), dtype=self.model.dtype, device=self.model.device)
        }
        current['target_val'] = self.model.log_target(current['sample'].clone().detach(), x, y)
        current['hamiltonian'] = self.hamiltonian(-current['target_val'], current['momentum'])

        proposed = {}
        proposed['sample'], proposed['momentum'], proposed['target_val'], _ = \
            self.leapfrog(current['sample'], current['momentum'], x, y)
        proposed['hamiltonian'] = self.hamiltonian(-proposed['target_val'], proposed['momentum'])

        ratio = torch.exp(current['hamiltonian'] - proposed['hamiltonian'])

        a = 2 * (ratio > 0.5) - 1

        while torch.pow(ratio, a) > torch.pow(2, -a):
            try:
                x, y = next(iterator)
            except StopIteration:
                iterator = iter(self.dataloader)
                x, y = next(iterator)

            self.step = (torch.pow(2, a) * self.step).item()

            current['target_val'] = self.model.log_target(current['sample'].clone().detach(), x, y)
            current['hamiltonian'] = self.hamiltonian(-current['target_val'], current['momentum'])

            _, proposed['momentum'], proposed['target_val'], _ = \
                self.leapfrog(current['sample'], current['momentum'], x, y)
            proposed['hamiltonian'] = self.hamiltonian(-proposed['target_val'], proposed['momentum'])

            ratio = torch.exp(current['hamiltonian'] - proposed['hamiltonian'])

    def set_current(self, theta, data=None):
        x, y = super().set_current(theta, data=data)
        self.current['target_val'], self.current['grad_val'] = \
            self.model.upto_grad_log_target(self.current['sample'].clone().detach(), x, y)

    def potential_energy(self, position, x, y):
        return -self.model.log_target(position, x, y)

    def upto_grad_potential_energy(self, position, x, y):
        target_val, grad_val = self.model.upto_grad_log_target(position, x, y)
        return -target_val, -grad_val # potential = -target_val, grad_potential = -grad_val

    def log_proposal(self, momentum):
        return - 0.5 * torch.sum(momentum**2)

    def kinetic_energy(self, momentum):
        return -self.log_proposal(momentum)

    def hamiltonian(self, potential, momentum):
        return potential + self.kinetic_energy(momentum)

    def leapfrog(self, position0, momentum0, x, y):
        position = position0.clone().detach()

        # Make a half step for momentum at the beginning
        potential, grad_potential = self.upto_grad_potential_energy(position, x, y)
        momentum = momentum0 - 0.5 * self.step * grad_potential

        # Alternate full steps for position and momentum
        for i in range(self.num_steps-1):
            # Make a full step for the position
            position = position + self.step * momentum

            # Make a full step for the momentum
            potential, grad_potential = self.upto_grad_potential_energy(position.clone().detach(), x, y)
            momentum = momentum - self.step * grad_potential

        # Make a half step for momentum at the end
        position = position + self.step * momentum
        potential, grad_potential = self.upto_grad_potential_energy(position.clone().detach(), x, y)
        momentum = momentum - 0.5 * self.step * grad_potential

        # Negate momentum at end of trajectory to make the proposal symmetric
        momentum = -momentum

        return position, momentum, -potential, -grad_potential # target_val = -potential, grad_val = -grad_potential

    def draw(self, x, y, savestate=False):
        proposed = {key : None for key in self.keys}

        proposed['sample'] = self.current['sample'].clone().detach()
        proposed['momentum'] = torch.randn(self.model.num_params(), dtype=self.model.dtype, device=self.model.device)

        self.current['momentum'] = proposed['momentum'].clone().detach()
        self.current['hamiltonian'] = self.hamiltonian(-self.current['target_val'], self.current['momentum'])

        proposed['sample'], proposed['momentum'], proposed['target_val'], proposed['grad_val'] = \
            self.leapfrog(proposed['sample'], proposed['momentum'], x, y)
        proposed['hamiltonian'] = self.hamiltonian(-proposed['target_val'], proposed['momentum'])

        rate = torch.min(
            torch.exp(self.current['hamiltonian'] - proposed['hamiltonian']),
            torch.tensor([1.], dtype=self.model.dtype, device=self.model.device)
        )

        if torch.rand(1, dtype=self.model.dtype, device=self.model.device) < rate:
            self.current['sample'] = proposed['sample'].clone().detach()
            self.current['target_val'] = proposed['target_val'].clone().detach()
            self.current['grad_val'] = proposed['grad_val'].clone().detach()
            self.current['accepted'] = 1
        else:
            self.model.set_params(self.current['sample'].clone().detach())
            self.current['accepted'] = 0

        if self.tuner is not None:
            if isinstance(self.tuner, HMCDATuner):
                if self.counter.idx < self.counter.num_burnin_iters:
                    self.step, self.num_steps = self.tuner.tune(
                        rate.item(), self.counter.idx, return_e=self.counter.idx != self.counter.num_burnin_iters - 1
                    )

        if savestate:
            self.chain.detach_and_update(self.current)

        self.current['sample'].detach_()
        self.current['target_val'].detach_()
        self.current['grad_val'].detach_()



from pathlib import Path

from .serial_sampler import SerialSampler

class SingleChainSerialSampler(SerialSampler):
    """ Serial MCMC Sampler with single chain"""
    def __init__(self, counter):
        super().__init__(counter=counter)

    def get_model(self):
        return self.model

    def get_chain(self):
        return self.chain

    def get_param(self, idx):
        return self.get_chain().get_param(idx)

    def get_sample(self, idx):
        return self.get_chain().get_sample(idx)

    def set_current(self, theta, data=None):
        self.current = {key : None for key in self.keys}
        self.current['sample'] = theta

        x, y = data or next(iter(self.dataloader))

        return x, y

    def set_all(self, theta, data=None):
        self.set_current(theta, data=data)

    def reset(self, theta, data=None, reset_counter=True, reset_chain=True):
        if reset_counter:
            self.counter.reset()
        if reset_chain:
            self.chain.reset(keys=self.chain.vals.keys())
        self.set_all(theta, data=data)

    def to_chainfile(self, path=Path.cwd(), mode='a'):
        self.chain.to_chainfile(path=path, mode=mode)



from .am import AM
from .demc import DEMC
from .gamc import GAMC
from .hmc import HMC
from .mala import MALA
from .metropolis_hastings import MetropolisHastings
from .multi_chain_serial_sampler import MultiChainSerialSampler
from .power_posterior_sampler import PowerPosteriorSampler
from .ram import RAM
from .sampler import Sampler
from .serial_sampler import SerialSampler
from .single_chain_serial_sampler import SingleChainSerialSampler
from .smmala import SMMALA



import torch
import torch.nn as nn

from torch.distributions import Normal

from .bayesian_model import BayesianModel

class Hyperparameters:
    def __init__(self, input_size=1, output_size=1, bias=True, activation=torch.sigmoid):
        self.input_size = input_size
        self.output_size = output_size
        self.bias = bias
        self.activation = activation

class LogisticRegression(BayesianModel):
    def __init__(self, loss, constraint=None, bounds=[-float('inf'), float('inf')], temperature=None, prior=None,
    hparams=Hyperparameters(), savefile=None, dtype=torch.float64, device='cpu'):
        super().__init__(
            loss=loss, constraint=constraint, bounds=bounds, temperature=temperature, dtype=dtype, device=device)
        self.hp = hparams
        self.linear = nn.Linear(self.hp.input_size, self.hp.output_size, bias=self.hp.bias).to(
            dtype=self.dtype, device=self.device
        )
        self.prior = prior or self.default_prior()
        if savefile:
            self.load_state_dict(savefile, strict=False)

    def default_prior(self):
        return Normal(
            torch.zeros(self.num_params(), dtype=self.dtype, device=self.device),
            torch.ones(self.num_params(), dtype=self.dtype, device=self.device)
        )

    def forward(self, x):
        x = self.linear(x)
        if self.hp.activation is not None:
            x = self.hp.activation(x)
        return x



import torch

from .homogeneous_kernel import HomogeneousKernel

class PeriodicKernel(HomogeneousKernel):
    """ Periodic kernel """

    def __init__(self, scale=1., l=1., p=2.):
        self.scale = scale # square of amplitude
        self.l = l # square or lengthscale
        self.p = p # multiple of period

    def k(self, x1, x2):
        return torch.exp(-torch.sin(self.dist(x1, x2).div(self.p)).pow(2).mul(2.).div(self.l)).mul(self.scale)



import torch

from torch.autograd import grad

from .model import Model

class LogTargetModel(Model):
    def __init__(self,
        constraint=None, bounds=[-float('inf'), float('inf')], temperature=None, dtype=torch.float64, device='cpu'):
        super().__init__(dtype=dtype, device=device)
        self.constraint = constraint
        self.bounds = bounds
        self.temperature = temperature

    def transform_params_forwards(self, theta):
        if (self.bounds[0] != -float('inf')) and (self.bounds[1] == float('inf')):
            theta_transformed = torch.log(theta - self.bounds[0])
        elif (self.bounds[0] == -float('inf')) and (self.bounds[1] != float('inf')):
            theta_transformed = torch.log(self.bounds[1] - theta)
        elif (self.bounds[0] != -float('inf')) and (self.bounds[1] != float('inf')):
            theta_transformed = - torch.log((self.bounds[1] - theta) / (theta - self.bounds[0]))
        return theta_transformed

    def transform_params_backwards(self, theta):
        if (self.bounds[0] != -float('inf')) and (self.bounds[1] == float('inf')):
            theta_transformed = torch.exp(theta) + self.bounds[0]
        elif (self.bounds[0] == -float('inf')) and (self.bounds[1] != float('inf')):
            theta_transformed = self.bounds[1] - torch.exp(theta)
        elif (self.bounds[0] != -float('inf')) and (self.bounds[1] != float('inf')):
            theta_transformed = (self.bounds[1] - self.bounds[0]) / (1 + torch.exp(-theta)) + self.bounds[0]
        return theta_transformed

    def transform_params(self, theta):
        if self.constraint == 'transformation':
            theta_transformed = self.transform_params_backwards(theta)
        elif self.constraint == 'truncation':
            theta_transformed = self.transform_params_forwards(theta)

        self.set_params(theta_transformed.clone().detach())

    def log_target(self, theta, x, y):
        raise NotImplementedError

    def grad_log_target(self, log_target_val):
        grad_log_target_val = grad(log_target_val, self.parameters(), create_graph=True)
        grad_log_target_val = torch.cat([g.view(-1) for g in grad_log_target_val])
        return grad_log_target_val

    def upto_grad_log_target(self, theta, x, y):
        log_target_val = self.log_target(theta, x, y)
        grad_log_target_val = self.grad_log_target(log_target_val)
        return log_target_val, grad_log_target_val

    def hess_log_target(self, grad_log_target_val):
        num_params = self.num_params()

        hess_log_target_val = []
        for i in range(num_params):
            deriv_i_wrt_grad = grad(grad_log_target_val[i], self.parameters(), retain_graph=True)
            hess_log_target_val.append(torch.cat([h.view(-1) for h in deriv_i_wrt_grad]))
        hess_log_target_val = torch.cat(hess_log_target_val, 0).reshape(num_params, num_params)

        return hess_log_target_val

    def metric_log_target(self, grad_log_target_val):
        return -self.hess_log_target(grad_log_target_val)

    def upto_hess_log_target(self, theta, x, y):
        log_target_val, grad_log_target_val = self.upto_grad_log_target(theta, x, y)
        hess_log_target_val = self.hess_log_target(grad_log_target_val)

        return log_target_val, grad_log_target_val, hess_log_target_val

    def upto_metric_log_target(self, theta, x, y):
        log_target_val, grad_log_target_val, hess_log_target_val = self.upto_hess_log_target(theta, x, y)
        return log_target_val, grad_log_target_val, -hess_log_target_val



import numpy as np
import torch
import torch.nn as nn

from eeyore.stats import binary_cross_entropy

torch_to_np_types = {torch.float32: np.float32, torch.float64: np.float64}

# Built-in function for binary classification
# https://github.com/pytorch/pytorch/issues/18945
# Second order automatic differentiation does not work after the pytorch issue has been merged
# import torch.nn.functional as F
# lambda x, y: F.binary_cross_entropy(x, y, reduction='sum')

loss_functions = {
    'binary_classification': lambda x, y: binary_cross_entropy(x, y, reduction='sum'),
    'multiclass_classification': lambda x, y: nn.CrossEntropyLoss(reduction='sum')(x, torch.argmax(y, 1))
}



from torch.distributions import MultivariateNormal

from .normalized_kernel import NormalizedKernel

class MultivariateNormalKernel(NormalizedKernel):
    """ Multivariate normal kernel """

    def __init__(self, loc, scale_tril):
        self.set_density(loc, scale_tril)

    def set_density(self, loc, scale_tril):
        """ Set multivariate normal probability density function """
        self.density = MultivariateNormal(loc, scale_tril=scale_tril)

    def set_density_params(self, loc, scale_tril=None):
        """ Set the parameters of multivariate normal probability density function """
        self.density.loc = loc
        if scale_tril is not None:
            self.density.scale_tril = scale_tril

    def k(self, x1, x2, scale=None):
        self.set_density_params(x2, scale=scale)
        return self.log_prob(x1).exp()



import copy
import itertools
import numpy as np
import torch

from pathlib import Path
from torch.distributions import Categorical

from .gamc import GAMC
from .mala import MALA
from .metropolis_hastings import MetropolisHastings
from .multi_chain_serial_sampler import MultiChainSerialSampler
from .smmala import SMMALA
from eeyore.chains import ChainFile, ChainList
from eeyore.datasets import DataCounter

class PowerPosteriorSampler(MultiChainSerialSampler):
    def __init__(self, model, dataloader, samplers,
        theta0=None, data0=None, counter=None,
        temperature=None, between_step=10, b=0.5, storage='list', keys=['sample', 'target_val', 'accepted'],
        path=Path.cwd(), mode='a', check_input=False):
        super(PowerPosteriorSampler, self).__init__(counter or DataCounter.from_dataloader(dataloader))
        self.between_step = between_step
        self.b = b

        self.num_chains = len(samplers)

        self.dataloader = dataloader

        self.sampler_names = [samplers[i][0] for i in range(self.num_chains)]

        self.use_gamc = all(name == 'GAMC' for name in self.sampler_names)
        if any(name == 'GAMC' for name in self.sampler_names) and (not self.use_gamc):
            raise ValueError

        self.init_samplers(model, samplers, theta0, data0 or next(iter(self.dataloader)), storage, keys, path, mode)

        self.set_temperature(temperature)

        self.dtype = self.samplers[0].samplers[0].model.dtype if self.use_gamc else self.samplers[0].model.dtype
        self.device = self.samplers[0].samplers[0].model.device if self.use_gamc else self.samplers[0].model.device

        if check_input:
            self.check_dtype()
            self.check_device()

        self.categoricals = []
        for i in range(self.num_chains):
            self.categoricals.append(Categorical(self.eval_categorical_probs(i)))

    def check_dtype(self):
        if self.use_gamc:
            valid = all(model.dtype == self.dtype for model in itertools.chain.from_iterable(
                [[sampler.samplers[i].model for i in range(2)] for sampler in self.samplers]
            ))
        else:
            valid = all(t == self.dtype for t in [sampler.model.dtype for sampler in self.samplers])

        if not valid:
            raise ValueError

    def check_device(self):
        if self.use_gamc:
            valid = all(model.device == self.device for model in itertools.chain.from_iterable(
                [[sampler.samplers[i].model for i in range(2)] for sampler in self.samplers]
            ))
        else:
            valid = all(t == self.device for t in [sampler.model.device for sampler in self.samplers])

        if not valid:
            raise ValueError

    def init_chain(self, i, storage, keys, path, mode):
        if storage == 'list':
            chain = ChainList(keys=keys)
        elif storage == 'file':
            chain_path = path.joinpath('chain'+f"{(i+1):0{len(str(self.num_chains))}}")
            if not chain_path.exists():
                chain_path.mkdir(parents=True, exist_ok=True)
            chain = ChainFile(keys=keys, path=chain_path, mode=mode)

        return chain

    def init_samplers(self, model, samplers, theta0, data0, storage, keys, path, mode):
        self.samplers = []
        for i in range(self.num_chains):
            if samplers[i][0] == 'MetropolisHastings':
                self.samplers.append(MetropolisHastings(
                    copy.deepcopy(model),
                    theta0=theta0, dataloader=None, data0=data0, counter=self.counter,
                    chain=self.init_chain(i, storage, keys, path, mode), **(samplers[i][1])
                ))
            elif samplers[i][0] == 'MALA':
                self.samplers.append(MALA(
                    copy.deepcopy(model),
                    theta0=theta0, dataloader=None, data0=data0, counter=self.counter,
                    chain=self.init_chain(i, storage, keys, path, mode), **(samplers[i][1])
                ))
            elif samplers[i][0] == 'SMMALA':
                self.samplers.append(SMMALA(
                    copy.deepcopy(model),
                    theta0=theta0, dataloader=None, data0=data0, counter=self.counter,
                    chain=self.init_chain(i, storage, keys, path, mode), **(samplers[i][1])
                ))
            elif samplers[i][0] == 'GAMC':
                self.samplers.append(GAMC(
                    copy.deepcopy(model),
                    theta0=theta0, dataloader=None, data0=data0, counter=self.counter,
                    chain=self.init_chain(i, storage, keys, path, mode), **(samplers[i][1])
                ))

    def default_indicator(self):
        return  self.num_chains - 1

    def set_temperature(self, temperature):
        if (temperature is not None) and (self.num_chains != len(temperature)):
            raise ValueError

        if (temperature is None):
            self.temperature = [(i/self.num_chains)**4 for i in range(1, self.num_chains+1)]
        else:
            self.temperature = temperature

        for i in range(self.num_chains):
            if self.use_gamc:
                for j in range(2):
                    self.samplers[i].samplers[j].model.temperature = self.temperature[i]
            else:
                self.samplers[i].model.temperature = self.temperature[i]

    def from_seq_to_events(self, k, i):
        j = k if (k < i) else (k+1)
        return j

    def from_events_to_seq(self, j, i):
        k = j if (j < i) else (j-1)
        return k

    def eval_categorical_prob(self, j, i):
        eb = np.exp(-self.b)
        numerator = eb**np.absolute(j-i)
        denominator = eb*(2-eb**i-eb**(self.num_chains-1-i))/(1-eb)
        return numerator/denominator

    def eval_categorical_probs(self, i):
        return torch.tensor(
            [self.eval_categorical_prob(j, i) for j in itertools.chain(range(i), range(i+1, self.num_chains))]
        )

    def categorical_log_prob(self, j, i):
        return self.categoricals[i].log_prob(torch.tensor(self.from_events_to_seq(j, i)))

    def sample_categorical(self, i):
        return self.from_seq_to_events(self.categoricals[i].sample().item(), i)

    def reset(self, theta, data=None):
        super().reset(theta, data=data, reset_counter=False, reset_chain=True)
        self.counter.reset()

    def within_chain_move(self, i, x, y):
        if self.sampler_names[i] == 'GAMC':
            self.samplers[i].reset_in_sampler_from_data(x, y, reset_counter=False, reset_chain=False)
        self.samplers[i].draw(x, y, savestate=False)

    def within_chain_moves(self, x, y):
        for i in range(self.num_chains):
            self.within_chain_move(i, x, y)

    def between_chain_move_log_rate(self, i, j, sampler_i, sampler_j, x, y):
        return self.categorical_log_prob(i, j) - \
            self.categorical_log_prob(j, i) - \
            sampler_i.current['target_val'] - \
            sampler_j.current['target_val'] + \
            sampler_i.model.log_target(sampler_j.current['sample'].clone().detach(), x, y) + \
            sampler_j.model.log_target(sampler_i.current['sample'].clone().detach(), x, y)

    def swap_states(self, i, j, x, y):
        if self.use_gamc:
            state_copy = copy.deepcopy(
                self.samplers[i].samplers[self.samplers[i].current_kernel].current['sample'].clone().detach()
            )
            self.samplers[i].reset_in_sampler(
                self.samplers[j].samplers[self.samplers[j].current_kernel].current['sample'].clone().detach(),
                data=(x, y),
                sampler_id=self.samplers[j].current_kernel,
                reset_counter=False,
                reset_chain=False
            )
            self.samplers[j].reset_in_sampler(
                state_copy.clone().detach(),
                data=(x, y),
                sampler_id=self.samplers[i].current_kernel,
                reset_counter=False,
                reset_chain=False
            )

            indicator_copy = self.samplers[i].current_kernel
            self.samplers[i].current_kernel = self.samplers[j].current_kernel
            self.samplers[j].current_kernel = indicator_copy
        else:
            state_copy = copy.deepcopy(self.samplers[i].current['sample'].clone().detach())
            self.samplers[i].reset(
                self.samplers[j].current['sample'].clone().detach(),
                data=(x, y),
                reset_counter=False,
                reset_chain=False
            )
            self.samplers[j].reset(state_copy.clone().detach(), data=(x, y), reset_counter=False, reset_chain=False)

    def revert_states(self, sampler_i, sampler_j):
        sampler_i.model.set_params(sampler_i.current['sample'].clone().detach())
        sampler_j.model.set_params(sampler_j.current['sample'].clone().detach())

    def between_chain_move(self, i, j, x, y):
        if self.use_gamc:
            log_rate = self.between_chain_move_log_rate(
                i, j, self.samplers[i].current_sampler(), self.samplers[j].current_sampler(), x, y
            )
        else:
            log_rate = self.between_chain_move_log_rate(i, j, self.samplers[i], self.samplers[j], x, y)

        if torch.log(torch.rand(1, dtype=self.dtype, device=self.device)) < log_rate:
            self.swap_states(i, j, x, y)
        else:
            if self.use_gamc:
                self.revert_states(self.samplers[i].current_sampler(), self.samplers[j].current_sampler())
            else:
                self.revert_states(self.samplers[i], self.samplers[j])

    def between_chain_moves(self, x, y):
        for i in range(self.num_chains):
            j = self.sample_categorical(i)

            self.between_chain_move(i, j, x, y)

    def save_state(self, i):
        if self.use_gamc:
            self.samplers[i].chain.detach_and_update(self.samplers[i].current_sampler().current)
        else:
            self.samplers[i].chain.detach_and_update(self.samplers[i].current)

    def draw(self, x, y, savestate=False):
        self.within_chain_moves(x, y)

        if ((self.counter.idx % self.between_step) == 0):
            self.between_chain_moves(x, y)

        if savestate:
            for i in range(self.num_chains):
                self.save_state(i)



import torch

from math import floor

def choose(n):
    return floor(torch.rand(1).item() * n)

def choose_from_subset(n, exclude):
    result = choose(n)

    while (result in exclude):
         result = choose(n)

    return result



from .cor_from_cov import cor_from_cov
from .mc_cov import mc_cov

def mc_cor(x, method='inse', adjust=False, rowvar=False):
    return cor_from_cov(mc_cov(x, method=method, adjust=adjust, rowvar=rowvar))



from .xydataset import XYDataset

class XYIDataset(XYDataset):
    def __init__(self, x, y):
        super().__init__(x, y)

    def __repr__(self):
        return f'XYIDataset: indexed XYDataset'

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx], idx

    @classmethod
    def from_xydataset(selfclass, xydataset):
        return selfclass(xydataset.x, xydataset.y)



import torch

from .single_chain_serial_sampler import SingleChainSerialSampler
from eeyore.chains import ChainList
from eeyore.datasets import DataCounter
from eeyore.stats import recursive_mean

class AM(SingleChainSerialSampler):
    def __init__(self, model,
        theta0=None, dataloader=None, data0=None, counter=None,
        cov0=None, l=0.05, b=1., c=1., t0=2, transform=None,
        chain=ChainList()):
        super(AM, self).__init__(counter or DataCounter.from_dataloader(dataloader))
        self.model = model
        self.dataloader = dataloader
        self.l = l
        self.b = b
        self.c = c
        self.t0 = t0
        self.transform = transform

        if cov0 is not None:
            self.cov0 = cov0.clone().detach()
        else:
            self.cov0 = torch.eye(self.model.num_params(), dtype=self.model.dtype, device=self.model.device)
        if self.transform is not None:
            self.cov0 = self.transform(self.cov0)
        self.num_accepted = 0
        self.keys = ['sample', 'target_val', 'accepted']
        self.chain = chain

        if theta0 is not None:
            self.set_all(theta0.clone().detach(), data=data0)

    def set_current(self, theta, data=None):
        x, y = super().set_current(theta, data=data)
        self.current['target_val'] = self.model.log_target(self.current['sample'].clone().detach(), x, y)

    def set_cov(self, cov=None):
        if cov is not None:
            self.cov = cov
        else:
            self.cov = self.cov0.clone().detach()
        self.running_mean = torch.zeros(self.model.num_params(), dtype=self.model.dtype, device=self.model.device)
        self.cov_sum = torch.zeros(
            self.model.num_params(), self.model.num_params(), dtype=self.model.dtype, device=self.model.device
        )

    def set_all(self, theta, data=None, cov=None):
        super().set_all(theta, data=data)
        self.set_cov(cov=cov)

    def reset(self, theta, data=None, reset_counter=True, reset_chain=True):
        super().reset(theta, data=data, reset_counter=reset_counter, reset_chain=reset_chain)
        self.num_accepted = 0

    def set_recursive_cov(self, n, offset=0):
        k = n - offset
        self.cov = (self.cov_sum - (k + 1) * torch.ger(self.running_mean, self.running_mean)) / k

    def draw(self, x, y, savestate=False, offset=0):
        proposed = {key : None for key in self.keys}

        randn_sample = torch.randn(self.model.num_params(), dtype=self.model.dtype, device=self.model.device)
        if (self.counter.idx + 1 - offset > self.t0):
            if torch.rand(1, dtype=self.model.dtype, device=self.model.device) < self.l:
                proposed['sample'] = self.current['sample'].clone().detach() + self.c * randn_sample
            else:
                proposed['sample'] = \
                    self.current['sample'].clone().detach() + self.b * torch.cholesky(self.cov) @ randn_sample
        else:
            proposed['sample'] = self.current['sample'].clone().detach() + self.c * randn_sample
        proposed['target_val'] = self.model.log_target(proposed['sample'].clone().detach(), x, y)

        log_rate = proposed['target_val'] - self.current['target_val']

        if torch.log(torch.rand(1, dtype=self.model.dtype, device=self.model.device)) < log_rate:
            self.current['sample'] = proposed['sample'].clone().detach()
            self.current['target_val'] = proposed['target_val'].clone().detach()
            self.current['accepted'] = 1
            if (self.counter.idx > 0):
                self.num_accepted = self.num_accepted + 1
        else:
            self.model.set_params(self.current['sample'].clone().detach())
            self.current['accepted'] = 0

        self.running_mean = recursive_mean(
            self.running_mean, self.counter.idx + 1, self.current['sample'], offset=offset
        )
        self.cov_sum = self.cov_sum + torch.ger(self.current['sample'], self.current['sample'])
        if (self.counter.idx + 1 - offset >= self.t0):
            if (self.num_accepted == 0):
                self.cov = self.cov0.clone().detach()
            else:
                self.set_recursive_cov(self.counter.idx, offset=offset)
                if self.transform is not None:
                    self.cov = self.transform(self.cov)

        if savestate:
            self.chain.detach_and_update(self.current)

        self.current['sample'].detach_()
        self.current['target_val'].detach_()



import torch

def running_mean(x, dim=0):
    if x.dim() == 1:
        return torch.cumsum(x, dim=0) / torch.arange(1, x.size(dim=0) + 1)
    elif x.dim() == 2:
        if dim == 0:
            return torch.cumsum(x, dim=0) / torch.arange(1, x.size(dim=0) + 1).view(-1, 1)
        elif dim == 1:
            return torch.cumsum(x, dim=1) / torch.arange(1, x.size(dim=1) + 1).view(1, -1)



from torch.utils.data import Dataset

class IDataset(Dataset):
    def __init__(self, dataset):
        self.dataset = dataset

    def __repr__(self):
        return f'IDataset: indexed Dataset'

    def __len__(self):
        return self.dataset.__len__()

    def __getitem__(self, idx):
        x, y = self.dataset.__getitem__(idx)
        return x, y, idx



from .demc_kernel import DEMCKernel
from .homogeneous_kernel import HomogeneousKernel
from .iso_se_kernel import IsoSEKernel
from .kernel import Kernel
from .multivariate_normal_kernel import MultivariateNormalKernel
from .normal_kernel import NormalKernel
from .normalized_kernel import NormalizedKernel
from .periodic_kernel import PeriodicKernel
from .rq_kernel import RQKernel
from .truncated_normal_kernel import TruncatedNormalKernel



# Implementation taken from
# https://stackoverflow.com/questions/43238173/python-convert-matrix-to-positive-semi-definite/43244194#43244194

import numpy as np
import torch

from .is_pos_def import is_pos_def

def nearest_pd(A, f=np.spacing):
    """Find the nearest positive-definite matrix to input

    A Python/Numpy port of John D'Errico's `nearestSPD` MATLAB code [1], which credits [2]

    [1] https://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd

    [2] https://doi.org/10.1016/0024-3795(88)90223-6
    """

    B = (A + A.T) / 2
    _, s, V = torch.svd(B)

    # For a comparison with kanga, see the following:
    # https://github.com/pytorch/pytorch/issues/16076#issuecomment-477755364
    H = torch.matmul(V, torch.matmul(torch.diag(s), V.T))

    A2 = (B + H) / 2

    A3 = (A2 + A2.T) / 2

    if is_pos_def(A3):
        return A3

    spacing = f(torch.norm(A).item())
    I = torch.eye(A.shape[0])
    k = 1
    while not is_pos_def(A3):
        eigenvals = torch.eig(A3, eigenvectors=False)[0][:, 0]
        mineig = eigenvals.min().item()
        A3 += I * (-mineig * k**2 + spacing)
        k += 1

    return A3



import torch

from .log_target_model import LogTargetModel
from eeyore.integrators import MCIntegrator

class BayesianModel(LogTargetModel):
    """ Class representing a Bayesian Net """
    def __init__(self, loss,
        constraint=None, bounds=[-float('inf'), float('inf')], temperature=None, dtype=torch.float64, device='cpu'):
        super().__init__(constraint=constraint, bounds=bounds, temperature=temperature, dtype=dtype, device=device)
        self.loss = loss

    def default_prior(self):
        """ Prior distribution """
        raise NotImplementedError

    def summary(self, hashsummary=False):
        print(self)
        print("-" * 80)
        n_params = self.num_params()
        print(f"Number of model parameters: {n_params}")
        print("-" * 80)
        print(f"Prior: {self.prior}")
        print("-" * 80)

        if hashsummary:
            print('Hash Summary:')
            for idx, hashvalue in enumerate(self.hashsummary()):
                print(f"{idx}: {hashvalue}")

    def log_lik(self, x, y):
        """ Log-likelihood """
        log_lik_val = -self.loss(self(x), y)
        if self.temperature is not None:
            log_lik_val = self.temperature * log_lik_val
        return log_lik_val

    def set_params_and_log_lik(self, theta, x, y):
        """ Set parameters and evaluate log-likelihood """
        if self.constraint is not None:
            self.transform_params(theta)
        else:
            self.set_params(theta)

        log_lik_val = self.log_lik(x, y)

        if self.constraint is not None:
            self.set_params(theta)

        return log_lik_val

    def log_prior(self):
        log_prior_val = torch.sum(self.prior.log_prob(self.get_params()))
        if self.temperature is not None:
            log_prior_val = self.temperature * log_prior_val
        return log_prior_val

    def log_target(self, theta, x, y):
        self.set_params(theta)

        log_prior_val = self.log_prior()

        if self.constraint is not None:
            self.transform_params(theta)

        log_lik_val = self.log_lik(x, y)

        if self.constraint is not None:
            self.set_params(theta)

        return log_lik_val + log_prior_val

    def predictive_posterior(self, theta, x, y): #, integration='mc'):
        integrator = MCIntegrator(f=self.set_params_and_log_lik, samples=theta) # if interation == 'mc':
        return integrator.integrate(x, y)

    def predictive_posterior_from_dataset(
        self, theta, dataset, num_points, shuffle=True, verbose=False, verbose_step=1):
        integrator = MCIntegrator(f=self.set_params_and_log_lik, samples=theta) # if interation == 'mc':
        return integrator.integrate_from_dataset(
            dataset, num_points, shuffle=shuffle, verbose=verbose, verbose_step=verbose_step
        )



import math
import torch
import torch.nn as nn

from torch.distributions import Normal, MultivariateNormal
from torchdiffeq import odeint

from eeyore.models import BayesianModel

class ODEModel(BayesianModel):
    def __init__(self, odes, eta, z0, noise_var=None, known_noise_var=False, constraint=None,
        bounds=[-float('inf'), float('inf')], temperature=None, prior=None, savefile=None, dtype=torch.float64,
        device='cpu'):
        super().__init__(
            loss=None, constraint=constraint, bounds=bounds, temperature=temperature, dtype=dtype, device=device
        )

        self.z0 = z0

        # eta are the ODE parameters and n_v is the noise variance
        self.eta = nn.Parameter(data=eta.to(dtype=self.dtype, device=self.device), requires_grad=True)

        if noise_var is None:
            self.n_v = torch.empty(self.num_states(), dtype=self.dtype, device=self.device)
        else:
            self.n_v = noise_var.to(dtype=self.dtype, device=self.device)

        if not known_noise_var:
            self.n_v = nn.Parameter(data=self.n_v, requires_grad=True)

        self.loss = self.columnwise_isotropic_normal_loss

        self.odes = odes

        self.prior = prior or self.default_prior()

    def columnwise_isotropic_normal_loss(self, z, y):
        return 0.5 * (
            torch.sum((y - z).pow(2).sum(0) / self.n_v)
            + self.num_times(z) * torch.sum(torch.log(self.n_v))
            + self.num_times(z) * self.num_states() * torch.log(torch.tensor(
                [2 * math.pi], dtype=self.dtype, device=self.device
            ))
        )

    def default_prior(self):
        return Normal(
            torch.zeros(self.num_params(), dtype=self.dtype, device=self.device),
            torch.ones(self.num_params(), dtype=self.dtype, device=self.device)
        )

    def num_states(self):
        return len(self.z0)

    def num_times(self, t):
        return t.shape[0]

    def forward(self, t):
        return odeint(lambda t, z: self.odes(t, z, self.eta), self.z0, t)

    def sample(self, t):
        z = self(t)

        y = torch.empty(self.num_times(t), self.num_states())
        for i in range(self.num_states()):
            y[:, i] = MultivariateNormal(
                    loc=z[:, i],
                    covariance_matrix=torch.diag(torch.tensor(
                        [self.n_v[i] for _ in range(self.num_times(t))], dtype=self.dtype, device=self.device)
                    )
                ).sample()

        return y, z



import torch

def recursive_cov(lastcov, lastmean, secondlastmean, n, x, offset=0):
    k = n - offset
    print("n =", n, ", offset =", offset, ", k =", k)
    return (
        (k - 1) * lastcov
        + torch.ger(x, x)
        - (k + 1) * torch.ger(lastmean, lastmean)
        + k * torch.ger(secondlastmean, secondlastmean)
        ) / k



class Sampler:
    """ Base class for sampling algorithms """

    def draw(self, x, y, savestate=False):
        raise NotImplementedError

    def run(self, num_epochs, num_burnin_epochs, verbose=False, verbose_step=100):
        raise NotImplementedError



from codecs import open
from os import path
from setuptools import find_packages, setup

from eeyore import __version__

url = 'https://github.com/papamarkou/eeyore'

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='eeyore',
    version=__version__,
    description='MCMC methods for neural networks',
    long_description=long_description,
    url=url,
    download_url='{0}/archive/v{1}.tar.gz'.format(url, __version__),
    packages=find_packages(),
    license='MIT',
    author='Theodore Papamarkou',
    author_email='theodore.papamarkou@gmail.com',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3'
    ],
    keywords=['Bayesian', 'deep learning', 'Markov chains', 'MCMC', 'Monte Carlo', 'neural networks'],
    python_requires='>=3.6',
    install_requires=['numpy>=1.19.2', 'torch>=1.6.0', 'torchdiffeq>=0.1.1', 'kanga>=0.0.17'],
    package_data={'eeyore': ['data/*/x.csv', 'data/*/y.csv', 'data/*/readme.md']},
    include_package_data=True,
    zip_safe=False
)



import torch

def is_pos_def(x):
    if torch.equal(x, x.t()):
        try:
            torch.cholesky(x)
            return True
        except RuntimeError:
            return False
    else:
        return False



import torch

from .single_chain_serial_sampler import SingleChainSerialSampler
from eeyore.chains import ChainList
from eeyore.datasets import DataCounter
from eeyore.kernels import NormalKernel

class MetropolisHastings(SingleChainSerialSampler):
    def __init__(self, model,
        theta0=None, dataloader=None, data0=None, counter=None,
        symmetric=True, kernel=None, chain=ChainList()):
        super(MetropolisHastings, self).__init__(counter or DataCounter.from_dataloader(dataloader))
        self.model = model
        self.dataloader = dataloader
        self.symmetric = symmetric
        if self.model.constraint == 'truncation':
            self.symmetric = False
        else:
            self.symmetric = symmetric

        self.keys = ['sample', 'target_val', 'accepted']
        self.chain = chain

        if theta0 is not None:
            self.set_current(theta0.clone().detach(), data=data0)

        self.kernel = kernel or self.default_kernel(self.current)

    def default_kernel(self, theta):
        return NormalKernel(theta, torch.ones(self.model.num_params()))

    def default_kernel(self, state):
        loc = state['sample']
        scale = torch.ones(self.model.num_params(), dtype=self.model.dtype, device=self.model.device)
        if (self.model.constraint is None) or (self.model.constraint == 'transformation'):
            return NormalKernel(loc, scale)
        elif self.model.constraint == 'truncation':
            return TruncatedNormalKernel(loc, scale, lower_bound=self.model.bounds[0], upper_bound=self.model.bounds[1])

    def set_current(self, theta, data=None):
        x, y = super().set_current(theta, data=data)
        self.current['target_val'] = self.model.log_target(self.current['sample'].clone().detach(), x, y)

    def reset(self, theta, data=None, reset_counter=True, reset_chain=True):
        super().reset(theta, data=data, reset_counter=reset_counter, reset_chain=reset_chain)
        self.set_kernel(self.current)

    def set_kernel(self, state, scale=None, scale_tril=None):
        self.kernel.set_density_params(state['sample'].clone().detach())

    def draw(self, x, y, savestate=False):
        proposed = {key : None for key in self.keys}

        proposed['sample'] = self.kernel.sample()
        proposed['target_val'] = self.model.log_target(proposed['sample'].clone().detach(), x, y)

        log_rate = proposed['target_val'] - self.current['target_val']
        if not self.symmetric:
            log_rate = log_rate - self.kernel.log_prob(proposed['sample'])
            self.set_kernel(proposed)
            log_rate = log_rate + self.kernel.log_prob(self.current['sample'])

        if torch.log(torch.rand(1, dtype=self.model.dtype, device=self.model.device)) < log_rate:
            self.current['sample'] = proposed['sample'].clone().detach()
            self.current['target_val'] = proposed['target_val'].clone().detach()
            if self.symmetric:
                self.set_kernel(proposed)
            self.current['accepted'] = 1
        else:
            self.model.set_params(self.current['sample'].clone().detach())
            if not self.symmetric:
                self.set_kernel(self.current)
            self.current['accepted'] = 0

        if savestate:
            self.chain.detach_and_update(self.current)

        self.current['sample'].detach_()
        self.current['target_val'].detach_()



import torch

from .homogeneous_kernel import HomogeneousKernel

class RQKernel(HomogeneousKernel):
    """ Rational quadratic kernel """

    def __init__(self, scale=1., l=1., a=1.):
        self.scale = scale # square of amplitude
        self.l = l # square of lengthscale
        self.a = a # scale mixture (a > 0)

    def k(self, x1, x2):
        return  self.squared_dist(x1, x2).div(2. * self.a * self.l).add(1.).pow(-self.a).mul(self.scale)



from .cov import cov
from .inse_mc_cov import inse_mc_cov

def mc_cov(x, method='inse', adjust=False, rowvar=False):
    if method == 'inse':
        return inse_mc_cov(x, adjust=adjust)
    elif method == 'iid':
        return cov(x, rowvar=rowvar)
    else:
        raise ValueError('The method can be inse or iid, {} was given'.format(method))



import copy
import numpy as np

from pathlib import Path

from .metropolis_hastings import MetropolisHastings
from .multi_chain_serial_sampler import MultiChainSerialSampler
from eeyore.chains import ChainFile, ChainList
from eeyore.datasets import DataCounter
from eeyore.kernels import DEMCKernel
from eeyore.stats import choose_from_subset

class DEMC(MultiChainSerialSampler):
    def __init__(self, model, scales, dataloader,
        theta0=None, data0=None, counter=None,
        num_chains=10, c=None, schedule=lambda n, num_iterations: 1., storage='list',
        keys=['sample', 'target_val', 'accepted'], path=Path.cwd(), mode='a'):
        super(DEMC, self).__init__(counter or DataCounter.from_dataloader(dataloader))
        self.scales = scales
        self.dataloader = dataloader
        self.num_chains = num_chains
        self.schedule = schedule

        self.c = c or [2.38/np.sqrt(model.num_params()) for i in range(self.num_chains)]

        self.init_samplers(model, theta0, data0 or next(iter(self.dataloader)), storage, keys, path, mode)

    def init_kernel(self, i, model):
        kernel = DEMCKernel(c=self.c[i])
        kernel.init_a_and_b(model.num_params(), model.dtype, model.device)
        kernel.init_density(model.num_params(), model.dtype, model.device)
        kernel.density.scale = self.scales[i]

        return kernel

    def init_chain(self, i, storage, keys, path, mode):
        if storage == 'list':
            chain = ChainList(keys=keys)
        elif storage == 'file':
            chain_path = path.joinpath('chain'+f"{(i+1):0{len(str(self.num_chains))}}")
            if not chain_path.exists():
                chain_path.mkdir(parents=True, exist_ok=True)
            chain = ChainFile(keys=keys, path=chain_path, mode=mode)

        return chain

    def init_samplers(self, model, theta0, data0, storage, keys, path, mode):
        self.samplers = []
        for i in range(self.num_chains):
            self.samplers.append(MetropolisHastings(
                copy.deepcopy(model),
                theta0=theta0, dataloader=None, data0=data0, counter=self.counter,
                symmetric=True, kernel=self.init_kernel(i, model), chain=self.init_chain(i, storage, keys, path, mode)
            ))

    def set_temperature(self, n, num_iterations):
        temperature = self.schedule(n, num_iterations)
        for i in range(self.num_chains):
            self.samplers[i].model.temperature = temperature

    def set_kernel(self, i):
        j = choose_from_subset(self.num_chains, [i])
        k = choose_from_subset(self.num_chains, [i, j])
        self.samplers[i].kernel.set_a_and_b(
            self.samplers[j].current['sample'].clone().detach(), self.samplers[k].current['sample'].clone().detach()
        )
        self.samplers[i].kernel.set_density_params(self.samplers[i].current['sample'].clone().detach())

    def reset(self, theta, data=None):
        super().reset(theta, data=data, reset_counter=False, reset_chain=True)
        self.counter.reset()

    def draw(self, x, y, savestate=False):
        self.set_temperature(self.counter.idx, self.counter.num_iters)

        for i in range(self.num_chains):
            self.set_kernel(i)
            self.samplers[i].draw(x, y, savestate=savestate)
